using Genie, Genie.Router, Genie.Renderer.Html, Genie.Requests
using HTTP
using Glob
using CommonMark
using TOML
using dg1d
import Base: Filesystem


function gather_outputdirs(root)
  pattern = "$root/*"
  dirs = glob(pattern)
  filter!(isdir , dirs)
  filter!(d->match(r".*\.previous",d) === nothing, dirs)
  return dirs
end


function html_twocolumn_scrollable(header, table, filenames_plots)

  parser = Parser()
  enable!(parser, TableRule())
  html_table = CommonMark.html(parser(table))

  html_plots = []
  for filename in filenames_plots
    !isfile(filename) && @warn "Path '$filename' is not a file"
    push!(html_plots, """<img src="/$(filename[10:end])" style="width:100%">""")
  end
  html_plots = join(html_plots, "\n")
  
  html_site = """
<!DOCTYPE html>
<html>
<style>

  .prev, .next, .home {
    cursor: pointer;
    position: absolute;
    color: white;
    font-weight: bold;
    font-size: 18px;
    user-select: none;
  }

  .home {
    left: 1%;
  }

  .next {
    right: 10%;
  }

  .prev {
    left: 10%;
  }

</style>
<body style="margin: 0; overflow: hidden">
  <div style="display: flex; flex-direction: column; height: 100vh">
    <div style="background-color: blue; 
                color: white; 
                display: flex;
                height: 50px;
                justify-content: center;
                align-items: center;">

    $(header)

    </div>
    <div style="display: flex; flex-direction: row; overflow: hidden">
      <div style="display: flex; flex-grow: 1; flex-basis: 30%; flex-direction: column; overflow-y: auto">

        <!-- table content goes here -->

        $(html_table)

      </div>
      <div style=" overflow-y: auto; display: flex; flex-basis: 70%; flex-direction: column;">

        <!-- plot content goes here -->

        $(html_plots...)

      </div>
    </div>
  </div>
</body>
</html>
"""

  return html_site
end


function to_markdown_table(dict::FlatDict)
  divider = [ "---" for _ in 1:length(dict) ]
  md = """
  | Parameter | Value |
  |---        |---    |
  """
  row_template(k, v) = "| $(k) | $(v) |\n"
  ks, vs = collect(keys(dict)), collect(values(dict))
  sort_indices = sortperm(ks)
  ks, vs = ks[sort_indices], vs[sort_indices]
  for (key, value) in zip(ks, vs)
    md *= row_template(key, value)
  end
  return md
end


function append_to(base_string, str)
  return join([base_string, str], "\n")
end


function header_template(route_name, id, dir; prev_id=missing, next_id=missing)

  link_prev = linkto(Symbol(route_name), sim_id=prev_id)
  link_next = linkto(Symbol(route_name), sim_id=next_id)

  header_home_button = """
  <a class="home" href="/">Home</a>
  """
  header_prev_button = """
  <a class="prev" href="../$(link_prev)">&#10094;</a>
  """
  header_next_button = """
  <a class="next" href="../$(link_next)">&#10095;</a>
  """
  header_text = """
  $(basename(dir))
  """

  header = header_home_button
  !ismissing(prev_id) && (header = append_to(header, header_prev_button))
  header = append_to(header, header_text)
  !ismissing(next_id) && (header = append_to(header, header_next_button))

  return header
end


function generate_html_sites(route_name, id, rootdir)

  table          = gather_parameters(rootdir)
  list_filenames = table."filename"
  all_ids        = map(list_filenames) do filename
    id_match = match(r"_(\d+).toml$", filename)
    isnothing(id_match) && error("Encountered filename with wrong pattern: '$filename'")
    return parse(Int64, first(id_match.captures))
  end
  !(id in all_ids) && error("Could not locate an output directory with number $id")
  p = sortperm(all_ids)
  table = table[p,:]
  all_ids = all_ids[p]
  list_filenames = list_filenames[p]
  list_outputdirs = map(list_filenames) do filename
    basename, _ = splitallext(filename)
    joinpath(rootdir, basename)
  end
  list_plotnames = [ glob("$dir/*.png") for dir in list_outputdirs ]
  parameters   = [ dg1d.to_dict(FlatDict, row) for row in eachrow(table) ]
  html_tables  = [ to_markdown_table(p) for p in parameters ]
  html_headers = map(list_outputdirs) do dir
    idx = findfirst(i -> i == id, all_ids)
    n_all_ids = length(all_ids)
    prev_id, next_id = if idx == 1
      all_ids[end], all_ids[2]
    elseif idx == n_all_ids
      all_ids[n_all_ids-1], all_ids[1]
    else
      all_ids[idx-1], all_ids[idx+1]
    end
    header_template(route_name, id, dir, prev_id=prev_id, next_id=next_id)
  end
  html_sites   = [ html_twocolumn_scrollable(header, tbl, plts)
                 for (header, tbl, plts) in zip(html_headers, html_tables, list_plotnames) ]
  return html_sites
end


function route_experiment(route_name, rootdir)
  sim_id = parse(Int64, payload(:sim_id))
  html_sites = generate_html_sites(route_name, sim_id, rootdir)
  return html_sites[sim_id]
end


function generate_routes_for_experiments()
  experiments_dir = "./public/experiments"
  root, dirs, _ = first(walkdir(experiments_dir))
  route_links = []
  route_names = []
  for dir in dirs
    name = Symbol(dir)
    route("$dir/:sim_id", () -> route_experiment(name, joinpath(root, dir)), named = name)
    link = "$dir/1"
    push!(route_links, link)
    push!(route_names, name)
  end

  return route_names, route_links
end


route("") do
  route_names, route_links = generate_routes_for_experiments()
  md_links = [ """ - [$name]($link)""" for (name, link) in zip(route_names, route_links) ]
  md_links = join(md_links, "\n")
  parser = Parser()
  html_links = CommonMark.html(parser(md_links))

  html_page = """
<!DOCTYPE html>
<html>
<style>
</style>
<body>

  $(html_links)

</body>
</html>
  """

  return html_page
end


up()
